const jwkToPem = require('jwk-to-pem');
const jwt = require('jsonwebtoken');

const TTL_MILLISECONDS = process.env.RPP_COGNITO_KEY_CACHE_TTL_MINUTES * 60 * 1000;
const COGNITO_SIGN_IN_URL = process.env.RPP_COGNITO_TOKEN_SIGNING_URL;
const COGNITO_ISSUER = (COGNITO_SIGN_IN_URL || '').replace('/.well-known/jwks.json', '');

const cache = {};

function invalidateCache(_cache) {
    const now = Date.now();
    const updatedAt = new Date(_cache.updatedAt || 0).getTime();
    if (now - updatedAt > TTL_MILLISECONDS) {
        _cache.keys = undefined;
        _cache.updatedAt = undefined;
    }
}

const fetcher = async (url) => {
    const res = await fetch(url);
    if (!res.ok) {
        throw new Error(`HTTP error! status: ${res.status}`);
    }
    return res.json();
};

async function getPublicKeys() {
    invalidateCache(cache);
    if (!cache.keys) {
        try {
            const data = await fetcher(COGNITO_SIGN_IN_URL);
            cache.keys = data.keys.reduce((memo, key) => {
                memo[key.kid] = key;
                return memo;
            }, {});
            cache.updatedAt = new Date().toISOString();
        } catch (error) {
            console.error('Failed to fetch AWS Cognito public keys:', error);
            throw new Error('Unable to fetch public keys');
        }
    }
    return cache.keys;
}

async function retryVerifyAwsToken(token, retryCount = 1) {
    const keys = await getPublicKeys();
    const decodedToken = jwt.decode(token, { complete: true });

    if (!decodedToken || !decodedToken.header.kid) {
        throw new Error('Invalid token');
    }

    const key = keys[decodedToken.header.kid];
    if (!key) {
        if (retryCount) {
            cache.keys = undefined;
            cache.updatedAt = undefined;
            return retryVerifyAwsToken(token, retryCount - 1);
        }
        throw new Error('Invalid token');
    }
    return key;
}

async function verifyAwsToken(token) {
    try {
        const key = await retryVerifyAwsToken(token);

        const publicKey = jwkToPem(key);

        return jwt.verify(token, publicKey, {
            issuer: COGNITO_ISSUER,
            algorithms: ['RS256'],
        });
    } catch (error) {
        console.error('Token verification failed:', error?.message);
        throw new Error('Unauthorized');
    }
}

module.exports = {
    verifyAwsToken
};
